import { resolve } from 'path';
import * as process from 'process';
import * as util from 'util';
import chalk from 'chalk';
import * as chokidar from 'chokidar';
import AggregateError from 'es-aggregate-error';
import _glob from 'glob';
import { emitGeneratedFiles } from './emitter/index.js';
import { Loader } from './loader/index.js';
import { createDefaultResolver } from './resolver/index.js';
import { createDefaultTransformer } from './transformer/index.js';
import { isMatchByGlob } from './util.js';
const glob = util.promisify(_glob);
export async function run(options) {
    const cwd = options.cwd ?? process.cwd();
    const silent = options.silent ?? false;
    const resolver = options.resolver ??
        createDefaultResolver({
            cwd,
            sassLoadPaths: options.sassLoadPaths,
            lessIncludePaths: options.lessIncludePaths,
            webpackResolveAlias: options.webpackResolveAlias,
        });
    const transformer = options.transformer ?? createDefaultTransformer({ cwd, postcssConfig: options.postcssConfig });
    const distOptions = options.outDir
        ? {
            rootDir: cwd,
            outDir: options.outDir,
        }
        : undefined;
    const loader = new Loader({ transformer, resolver });
    const isExternalFile = (filePath) => {
        return !isMatchByGlob(filePath, options.pattern, { cwd });
    };
    async function processFile(filePath) {
        try {
            const result = await loader.load(filePath);
            await emitGeneratedFiles({
                filePath,
                tokens: result.tokens,
                distOptions,
                emitDeclarationMap: options.declarationMap,
                dtsFormatOptions: {
                    localsConvention: options.localsConvention,
                },
                silent,
                cwd,
                isExternalFile,
            });
        }
        catch (error) {
            if (error instanceof Error) {
                // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                console.error(chalk.red('[Error] ' + error.stack));
            }
            else {
                // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                console.error(chalk.red('[Error] ' + error));
            }
            throw error;
        }
    }
    if (options.watch) {
        if (!silent)
            console.log('Watch ' + options.pattern + '...');
        const watcher = chokidar.watch([options.pattern.replace(/\\/g, '/')], { cwd });
        watcher.on('all', (eventName, filePath) => {
            if (eventName === 'add' || eventName === 'change') {
                processFile(resolve(cwd, filePath)).catch(() => {
                    // TODO: Emit a error by `Watcher#onerror`
                });
            }
        });
        return { close: async () => watcher.close() };
    }
    else {
        const filePaths = (await glob(options.pattern, { dot: true, cwd }))
            // convert relative path to absolute path
            .map((file) => resolve(cwd, file));
        // TODO: Use `@file-cache/core` to process only files that have changed
        const errors = [];
        for (const filePath of filePaths) {
            await processFile(filePath).catch((e) => errors.push(e));
        }
        if (errors.length > 0)
            throw new AggregateError(errors, 'Failed to process files');
    }
}
//# sourceMappingURL=runner.js.map